import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:flutter_live_data/index.dart';
import 'package:flutter_live_data/index.dart' as fld;
import 'life_cycle_listener.dart';
import 'component_view.dart';
import 'package:logger/logger.dart' as leisim;

class _ImComponentProperties {
  bool isInitializingState = false;
  bool isCreatingState = false;
  late Widget view;
  void Function()? setRebuild;
}

class LogicComponent extends StatefulWidget implements LifeCycleListener, LifeCycleOwner {
  final Widget Function(LogicComponent) builder;
  final List<LiveData> _liveDataList = <LiveData>[];
  final BlocComponentDependencies dependencies = BlocComponentDependencies();
  final _ImComponentProperties attr = _ImComponentProperties();

  final SubscribeLifeCycleEvent _subscribeLifeCycleEvent = SubscribeLifeCycleEvent();
  final PeriodicLifeCycleEvent _periodicLifeCycleEvent = PeriodicLifeCycleEvent();

  final leisim.Logger? logger;

  BuildContext? get context => dependencies.context;

  LogicComponent({
    Key? key,
    required this.builder,
    this.logger,
  }) : super(key: key) {
    attr.view = builder(this);

    Future.delayed(const Duration(microseconds: 1), () {
      attr.isCreatingState = true;
      construct();
      attr.isCreatingState = false;
    });
  }

  construct() {}

  rebuild() {
    attr.setRebuild?.call();
  }

  @override
  State<StatefulWidget> createState() => _LogicComponentState();

  @override
  void observeLiveData<T>(LiveData<T> liveData) {
    _liveDataList.add(liveData);
  }

  void clearObserveLiveData() {
    for (var liveData in _liveDataList) {
      liveData.close();
    }
    _liveDataList.clear();
  }

  Widget bindView(Widget view) {
    return attr.view = view;
  }

  // void bloComponentSelfInit() {}

  subscribe<T>(
      fld.LiveData<T>? liveData,
      void Function(T value) onData,
      ) {
    if (liveData == null) return;
    if (attr.isInitializingState) {
      logger?.w('call [subscribe] inside method onInit (initState) cause abnormal behavior when Hot-Reload'
          '\n'
          'hint: move code to class constructor or onCreate');
    }
    return _subscribeLifeCycleEvent.call(liveData, onData);
  }

  PeriodicTimerSnap periodic(
      Duration duration,
      void Function(Timer timer) callback,
      ) {
    if (attr.isInitializingState) {
      logger?.w('call [periodic] inside method onInit (initState) cause abnormal behavior when Hot-Reload'
          '\n'
          'hint: move code to class constructor or onCreate');
    }

    return _periodicLifeCycleEvent.call(duration, callback);
  }

  @override
  void onInit() {
    // TODO: implement onBuild
  }

  @override
  void onDispose() {
    // TODO: implement onBuild
  }

  @override
  void onPause() {
    // TODO: implement onPause
  }

  @override
  void onResume() {
    // TODO: implement onResume
  }
}

class _LogicComponentState extends State<LogicComponent> with WidgetsBindingObserver {
  @override
  void initState() {
    super.initState();
    widget.dependencies.context = context;
    Future.delayed(const Duration(microseconds: 1), () {
      widget.attr.isInitializingState = true;
      WidgetsBinding.instance.addObserver(this);
      widget.onInit();
      // widget.bloComponentSelfInit();
      _asView(widget.attr.view)?.onInit();
      widget.attr.isInitializingState = false;
    });
  }

  @override
  void dispose() {
    widget.onDispose();
    widget.clearObserveLiveData();
    _asView(widget.attr.view)?.onDispose();
    WidgetsBinding.instance.removeObserver(this);
    widget.attr.setRebuild = null;
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    widget.dependencies.context = context;
    widget.attr.setRebuild = rebuild;
    return widget.bindView(widget.builder(widget));
  }

  rebuild() {
    setState(() {});
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.resumed) {
      widget.onResume();
      _asView(widget.attr.view)?.onResume();
    } else if (state == AppLifecycleState.paused) {
      widget.onPause();
      _asView(widget.attr.view)?.onPause();
    }
  }
}

View? _asView(Widget w) => w is View ? w : null;

class BlocComponentDependencies {
  BuildContext? context;
}

class SubscribeLifeCycleEvent {
  final List<dynamic> streamSubscription = [];

  call<T>(
      fld.LiveData<T> liveData,
      void Function(T value) onData,
      ) {
    streamSubscription.add(liveData.listen(onData)!);
  }

  onConstruct() {}

  onInit() {}

  onPause() {
    for (var s in streamSubscription) {
      if (s is! StreamSubscription) continue;
      StreamSubscription subscription = s;
      subscription.pause();
    }
  }

  onResume() {
    for (var s in streamSubscription) {
      if (s is! StreamSubscription) continue;
      StreamSubscription subscription = s;
      subscription.resume();
    }
  }

  onDispose() {
    onPause();
    streamSubscription.clear();
  }
}

class PeriodicLifeCycleEvent {
  final List<PeriodicTimerSnap> periodicTimerSnap = <PeriodicTimerSnap>[];

  call(
      Duration duration,
      void Function(Timer timer) callback,
      ) {
    var snap = PeriodicTimerSnap(duration, callback);
    _startTask(snap);
    periodicTimerSnap.add(snap);
    return snap;
  }

  _startTask(PeriodicTimerSnap snap) {
    snap.timer = Timer.periodic(snap.duration, (Timer timer) {
      snap.callback(timer);
      snap.run = true;
    });
  }

  _cancel() {
    for (var snap in periodicTimerSnap) {
      snap.timer?.cancel();
      snap.run = false;
    }
  }

  onConstruct() {
    for (var snap in periodicTimerSnap) {
      if (!snap.run) {
        _startTask(snap);
      }
    }
  }

  onInit() {}

  onPause() {
    _cancel();
  }

  onResume() {
    for (var snap in periodicTimerSnap) {
      snap.timer = Timer.periodic(snap.duration, (Timer timer) {
        snap.callback(timer);
        snap.run = false;
      });
    }
  }

  onDispose() {
    _cancel();
    periodicTimerSnap.clear();
  }
}

class PeriodicTimerSnap {
  final Duration duration;
  final void Function(Timer timer) callback;
  Timer? timer;
  bool run = false;

  PeriodicTimerSnap(this.duration, this.callback);

  PeriodicTimerSnap startNow() {
    Future.delayed(const Duration(microseconds: 1), () {
      if (timer != null) {
        callback(timer!);
      }
    });
    return this;
  }

  @override
  String toString() {
    return 'PeriodicTimerSnap{duration: $duration, callback: $callback, timer: $timer, run: $run}';
  }
}
